from typing import Callable, Dict, List

import matplotlib.pyplot as plt
import numpy as np
import talib as ta
from crypto_data_downloader.utils import load_pkl
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize
from scipy.stats import binned_statistic
from trading_models.utils import D2_TYPE, D_TYPE, plot_general

from fast_trading_simulator.sim import ACT_HIGH, ACT_LOW, ACT_NAMES


class ActMap:
    @staticmethod
    def from_tanh(tanh, low, high):
        return (tanh + 1) / 2 * (high - low) + low

    @staticmethod
    def to_tanh(x, low, high):
        return (x - low) / (high - low) * 2 - 1


def volatility(price: np.ndarray):
    dp = np.diff(price) / price[:-1]
    return np.sqrt(np.mean(dp**2))


def make_market_n_obs(
    path,
    # vol_range=[1e-3, 2e-2],
    ref_sym="BTCUSDT",
    price_idx=1,
    # obs:
    MA=ta.KAMA,
    periods=[10, 100],
    add_ref_obs=True,
):
    data: D_TYPE = load_pkl(path, gz=True)
    T = len(data[ref_sym])
    temp: D2_TYPE = {}
    skip = max(periods)
    for sym, v in data.items():
        if len(v) != T:
            continue
        p = v[:, price_idx]
        obs = np.array([p / MA(p, P) - 1 for P in periods]).T
        obs = obs.clip(-1, 1)
        vol = volatility(p)
        # ok = vol > vol_range[0] and vol < vol_range[1]
        ok = vol > 0
        temp[sym] = {"raw": v[skip:], "obs": obs[skip:], "ok": ok, "vol": vol}
    if add_ref_obs:
        ref_obs = temp[ref_sym]["obs"]
        for sym, d in temp.items():
            d["obs"] = np.concat([d["obs"], ref_obs], axis=1)
    get = lambda key: np.array([d[key] for d in temp.values() if d["ok"]])
    # plt.hist(get("vol"), bins=100)
    # plt.savefig("volatility_hist")
    return get("raw"), get("obs")


def rand_period(arrays: List[np.ndarray], dt=1024):
    T = arrays[0].shape[1]
    t1 = np.random.randint(0, T - dt)
    return [x[:, t1 : t1 + dt] for x in arrays]


def rand_action(obs: np.ndarray):
    SYM, TIME, _ = obs.shape
    tanh_act = np.random.uniform(-1, 1, (SYM, TIME, 5))
    tanh_act[:, :, 0] = np.where(tanh_act[:, :, 0] > 0, 1, -1)
    return ActMap.from_tanh(tanh_act, ACT_LOW, ACT_HIGH)


def plot_act_hist(action: np.ndarray):
    plots = {f"{ACT_NAMES[i]}_hist": action[..., i] for i in range(5)}
    plot_general(plots, "act_hist")


def bin_stat(X: np.ndarray, Y: np.ndarray, stat, y_lim=None, bins=100):
    nx, ny = X.shape[-1], Y.shape[-1]
    plt.figure(figsize=(4 * ny, 3 * nx))
    cnt = 0
    for i in range(nx):
        for j in range(ny):
            cnt += 1
            plt.subplot(nx, ny, cnt)
            plt.title(f"X{i}, Y{j}")
            x, y = X[:, i], Y[:, j]
            y2, edges, _ = binned_statistic(x, y, stat, bins)
            x2 = (edges[:-1] + edges[1:]) / 2
            plt.scatter(x, y, s=3, c="y", label="raw")
            plt.plot(x2, y2, c="b", label="bin_stat")
            if y_lim is not None:
                plt.ylim(y_lim)
            plt.legend()
    plt.tight_layout()
    plt.savefig("bin_stat")
    plt.close()


# ======================================


def round_dx(x, dx):
    return round(round(x / dx) * dx, 10)


def pymoo_minimize(func: Callable, conf: Dict, algo=GA()):
    xl = [v[0] for v in conf.values()]
    xu = [v[1] for v in conf.values()]
    dx = [v[2] for v in conf.values()]
    best = np.inf

    class Prob(ElementwiseProblem):
        def __init__(s):
            super().__init__(n_var=len(xl), n_obj=1, xl=xl, xu=xu)

        def _evaluate(s, X: np.ndarray, out, *args, **kwargs):
            X = [round_dx(xi, dxi) for xi, dxi in zip(X, dx)]
            P = dict(zip(conf.keys(), X))
            loss = func(P)
            nonlocal best
            if loss < best:
                best = loss
                print(f"best: {best} {P}")
            out["F"] = loss

    minimize(Prob(), algo, seed=42)



import numpy as np
import talib as ta
from crypto_data_downloader.utils import load_pkl
from trading_models.utils import D_TYPE

from fast_trading_simulator.sim import map_trades, simulate

path = "./futures_data_2025-08-01_2025-11-20.pkl"
data: D_TYPE = load_pkl(path, gz=True)

ref_sym = "BTCUSDT"
price_idx = 1
delta = 0.05

T = len(data[ref_sym])
data = {k: v for k, v in data.items() if len(v) == T}
all_act = []
for sym, v in data.items():
    p = v[:, price_idx]
    obs = p / ta.KAMA(p, 20) - 1
    pos = np.where(obs < -delta, 1, 0)
    pos = np.where(obs > delta, -1, pos)
    lev = np.full(T, 20)
    timeout = np.full(T, 10)
    take_profit = np.full(T, 0.01)
    stop_loss = np.full(T, -0.9)
    act = np.array([pos, lev, timeout, take_profit, stop_loss]).T
    all_act.append(act)
market = np.array(list(data.values()))
action = np.array(all_act)
tot_fee = 1e-3
liq_fee = np.full(len(data), 0.02)

trades = simulate(market, action, tot_fee, liq_fee, use_ratio=0.2, alloc_ratio=0.01)
map_trades(trades, plot=True)



from typing import Dict, List

import numba
import numpy as np
from trading_models.utils import plot_general

E_NO_EXIT = 0
E_TIMEOUT = 1
E_TAKE_PROFIT = 2
E_STOP_LOSS = 3
E_LIQ_PROTECT = 4
E_LIQ_TRIGGER = 5

EXIT_REASONS = [
    "no_exit",
    "timeout",
    "take_profit",
    "stop_loss",
    "liq_protect",
    "liq_trigger",
]

ACT_NAMES = "pos, lev, timeout, take_profit, stop_loss".split(",")
ACT_LOW = np.array([-1, 1, 1, 0.01, -0.9])
ACT_HIGH = np.array([1, 10, 100, 0.2, -0.1])


@numba.njit
def find_profit(
    price1: float,
    price2: float,
    dt: int,
    action: np.ndarray,
    tot_fee: float,
    liq_fee: float,
    liq_protect=-0.8,
    liq_trigger=-0.9,
    clip_pr: bool = False,
):
    pos, lev, timeout, take_profit, stop_loss = action
    pos, lev, p1, p2 = np.sign(pos), max(1, int(lev)), price1, price2
    pr = lev * (pos * (p2 - p1) / p1 - tot_fee)
    liq_pr = pr - lev * (liq_fee * p2 / p1)
    if clip_pr:
        pr = min(pr, take_profit)
    if liq_pr < liq_trigger:
        return -1.0, E_LIQ_TRIGGER
    if liq_pr < liq_protect:
        return pr, E_LIQ_PROTECT
    if pr < stop_loss:
        return pr, E_STOP_LOSS
    if pr >= take_profit:
        return pr, E_TAKE_PROFIT
    if dt >= timeout:
        return pr, E_TIMEOUT
    return pr, E_NO_EXIT


@numba.njit
def find_price2(
    price1: float,
    action: np.ndarray,
    tot_fee: float,
):
    pos, lev, timeout, take_profit, stop_loss = action
    pos, lev, p1 = np.sign(pos), max(1, int(lev)), price1
    # pr = lev * (pos * (p2 - p1) / p1 - tot_fee)
    return (take_profit / lev + tot_fee) * p1 / pos + p1


@numba.njit
def simulate(
    market: np.ndarray,
    action: np.ndarray,
    tot_fee: float,
    liq_fee: np.ndarray,
    liq_protect=-0.8,
    liq_trigger=-0.9,
    clip_pr: bool = False,
    min_pos=0.1,
    use_ratio=0.5,
    alloc_ratio=0.01,
    init_cash=10e3,
    min_cash=10,
    price_idx=1,
):
    max_open = int(use_ratio / alloc_ratio)
    SYMBOL, TIME, _ = market.shape

    worth = cash_left = init_cash
    open_trades: Dict[int, np.ndarray] = {}
    done_trades: List[np.ndarray] = []

    for t in range(TIME):

        for id, x in list(open_trades.items()):
            s1, t1, cash = int(x[0]), int(x[1]), x[2]
            dt = t - t1
            pr, exit = find_profit(
                price1=market[s1, t1, price_idx],
                price2=market[s1, t, price_idx],
                dt=dt,
                action=action[s1, t1],
                tot_fee=tot_fee,
                liq_fee=liq_fee[s1],
                liq_protect=liq_protect,
                liq_trigger=liq_trigger,
                clip_pr=clip_pr,
            )
            if exit:
                worth += cash * pr
                cash_left += cash * (1 + pr)
                x[-4:] = float(dt), pr, float(exit), worth
                del open_trades[id]
                done_trades.append(x)

        for s in range(SYMBOL):
            pos = action[s, t, 0]
            if abs(pos) > min_pos and len(open_trades) < max_open:
                cash = min(cash_left, worth * alloc_ratio * abs(pos))
                id = (np.sign(pos), s)  # int(np.sign(pos) * (s + 1))
                if cash > min_cash and id not in open_trades:
                    cash_left -= cash
                    open_trades[id] = np.array([s, t, cash, pos, 0.0, 0.0, 0.0, 0.0])

    return done_trades


def map_trades(trades, to_map=[], plot=True):
    trades = np.array(trades).T
    sym, time = trades[:2].astype(int)
    keys = ["cash", "position", "duration", "profit", "exit", "worth"]
    types = ["_hist", "_hist", "_hist", "_hist", "_hist", ""]
    res: Dict[str, np.ndarray] = {k: v for k, v in zip(keys, trades[2:])}
    res["exit"] = np.array(EXIT_REASONS)[res["exit"].astype(int)]
    to_map = [v[sym, time] for v in to_map]
    if plot:
        plots = {f"{k}{type}": res[k] for k, type in zip(keys, types)}
        plots[f"log10(worth) ({len(sym)} trades)"] = np.log10(res["worth"])
        plot_general(plots, "simulate")
    return (res, to_map) if to_map else res


@numba.njit
def sim_all_points(
    market: np.ndarray,
    action: np.ndarray,
    tot_fee: float,
    liq_fee: np.ndarray,
    liq_protect=-0.8,
    liq_trigger=-0.9,
    clip_pr: bool = False,
    price_idx=1,
):
    SYMBOL, TIME, _ = market.shape
    all_pr = np.zeros((SYMBOL, TIME, 1))
    for s in numba.prange(SYMBOL):
        for t in numba.prange(TIME):
            act = action[s, t]
            if act[0] == 0:
                continue
            for t2 in range(t + 1, TIME):
                pr, exit = find_profit(
                    price1=market[s, t, price_idx],
                    price2=market[s, t2, price_idx],
                    dt=t2 - t,
                    action=act,
                    tot_fee=tot_fee,
                    liq_fee=liq_fee[s],
                    liq_protect=liq_protect,
                    liq_trigger=liq_trigger,
                    clip_pr=clip_pr,
                )
                if exit:
                    all_pr[s, t, 0] = pr
                    break
    return all_pr
